﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Collections;

namespace Persistence {

  [AttributeUsage(
    AttributeTargets.Field | AttributeTargets.Property,
    AllowMultiple=false,
    Inherited=false)]
  public class StorableAttribute : Attribute {

    private string name = null;
    private object defaultValue = null;

    public string Name {
      get { return this.name; }
      set { this.name = value; }
    }

    public object DefaultValue {
      get { return this.defaultValue; }
      set { this.defaultValue = value; }
    }  

    private const BindingFlags instanceMembers =
      BindingFlags.Instance |
      BindingFlags.Public |
      BindingFlags.NonPublic;

    public static Dictionary<string, DataMemberAccessor> GetAutostorableAccessors(object obj) {
      Dictionary<string, DataMemberAccessor> autoStorableAccessors = 
        new Dictionary<string, DataMemberAccessor>();
      foreach (MemberInfo memberInfo in obj.GetType().GetMembers(instanceMembers)) {
        foreach (object attribute in memberInfo.GetCustomAttributes(false)) {          
          StorableAttribute autoStorableAttribute =
            attribute as StorableAttribute;
          if (autoStorableAttribute != null) {
            autoStorableAccessors.Add(memberInfo.Name,
              new DataMemberAccessor(memberInfo, autoStorableAttribute, obj));
          }
        }
      }
      return autoStorableAccessors;
    }    
  }

  class CloningFactory {

    private static CloningFactory instance = new CloningFactory();

    public static T DefaultClone<T>(T obj) {
      return instance.Clone(obj);
    }

    private List<ICompoundSerializer> compoundSerializers;

    public CloningFactory(IEnumerable<ICompoundSerializer> compoundSerializers) {
      this.compoundSerializers = new List<ICompoundSerializer>(compoundSerializers);
    }

    public CloningFactory() :
      this(InterfaceInstantiatior.InstantiateAll<ICompoundSerializer>()) { }

    delegate object CloneMethod(object obj, Dictionary<object, object> twins);

    Dictionary<Type, CloneMethod> cloneMethods = new Dictionary<Type, CloneMethod>();

    public T Clone<T>(T obj) {
      Dictionary<object, object> twins = new Dictionary<object, object>();
      return Clone(obj, twins);
    }

    public T Clone<T>(T obj, Dictionary<object, object> twins) {
      if (obj == null)
        return default(T);

      ValueType t = obj as ValueType;
      if (t != null)
        return (T)obj;

      if (twins.ContainsKey(obj))
        return (T)twins[obj];

      ICloneable c = obj as ICloneable;
      if (c != null) {
        T newClone = (T)c.Clone();
        twins[obj] = newClone;
        return newClone;
      }

      Type type = obj.GetType();
      if (!cloneMethods.ContainsKey(type))
        cloneMethods[type] = CreateCloneMethod(type);
      return (T)cloneMethods[type](obj, twins);
    }

    private CloneMethod CreateCloneMethod(Type type) {
      if (type.GetConstructor(new Type[] { type }) != null)
        return CopyConstructorCloneMethod.Create(type);
      foreach (ICompoundSerializer serializer in compoundSerializers)
        if (serializer.CanSerialize(type))
          return CompoundSerializerCloneMethod.Create(serializer, type, this);
      return StorableAccessorCloneMethod.Create(type, this);      
    }

    class CopyConstructorCloneMethod {
      public static CloneMethod Create(Type type) {
        return new CopyConstructorCloneMethod(type).method;
      }
      private ConstructorInfo constructorInfo;
      public CopyConstructorCloneMethod(Type type) {
        this.constructorInfo = type.GetConstructor(new Type[] { type });
      }

      object method(object obj, Dictionary<object, object> twins) {
        object newInstance = constructorInfo.Invoke(new object[] { obj });
        twins.Add(obj, newInstance);
        return newInstance;
      }
    }

    class CompoundSerializerCloneMethod {
      public static CloneMethod Create(ICompoundSerializer serializer,
          Type type, CloningFactory factory) {
        return new CompoundSerializerCloneMethod(serializer, type, factory).method;
      }
      private ICompoundSerializer serializer;
      private Type type;
      private CloningFactory factory;
      public CompoundSerializerCloneMethod(ICompoundSerializer serializer,
          Type type, CloningFactory factory) {
        this.serializer = serializer;
        this.type = type;
        this.factory = factory;
      }
      object method(object obj, Dictionary<object, object> twins) {
        return serializer.DeSerialize(
          Functional.Map(
            serializer.Serialize(obj),
            (o) => factory.Clone(o, twins)),
          type);
      }
    }

    class StorableAccessorCloneMethod {

      delegate object Getter(object target);
      delegate void Setter(object target, object value);

      struct FieldAccessor {      
        public Getter Getter;
        public Setter Setter;
        public FieldAccessor(Getter getter, Setter setter) {
          this.Getter = getter;
          this.Setter = setter;
        }
      }

      private const BindingFlags INSTANCE_MEMBERS =
        BindingFlags.Instance |
        BindingFlags.Public |
        BindingFlags.NonPublic;

      public static CloneMethod Create(Type type, CloningFactory factory) {
        return new StorableAccessorCloneMethod(type, factory).method;
      }

      private List<FieldAccessor> fieldAccessors;
      private CloningFactory factory;
         
      public StorableAccessorCloneMethod(Type type, CloningFactory factory) {
        this.factory = factory;
        this.fieldAccessors = new List<FieldAccessor>();
        foreach (MemberInfo memberInfo in type.GetMembers(INSTANCE_MEMBERS)) {
          foreach (object attribute in memberInfo.GetCustomAttributes(false)) {
            StorableAttribute autoStorableAttribute =
              attribute as StorableAttribute;
            if (autoStorableAttribute != null) {
              if (memberInfo.MemberType == MemberTypes.Field) {
                FieldInfo fi = (FieldInfo)memberInfo;
                fieldAccessors.Add(new FieldAccessor(
                  fi.GetValue,
                  fi.SetValue));
              } else if (memberInfo.MemberType == MemberTypes.Property) {
                PropertyInfo pi = (PropertyInfo)memberInfo;
                fieldAccessors.Add(new FieldAccessor(
                  (target) => pi.GetValue(target, null),
                  (target, value) => pi.SetValue(target, value, null)));
              }
            }
          }
        }
      }
      object method(object obj, Dictionary<object, object> twins) {
        object newInstance = Activator.CreateInstance(obj.GetType());
        twins.Add(obj, newInstance);
        foreach (FieldAccessor fa in this.fieldAccessors) {
          fa.Setter(newInstance, factory.Clone(fa.Getter(obj), twins));
        }
        return newInstance;
      }
    }
  }
  
  public interface IDataMemberAccessor {
    string Name { get; }
    Type Type { get; }
    object DefaultValue { get; }
    object Get();
    void Set(object value);    
  }
    
  public class DataMemberAccessor : IDataMemberAccessor {

    public delegate object Getter();
    public delegate void Setter(object value);

    private string name;
    private Type type;
    private object defaultValue;
    private Getter getter;
    private Setter setter;

    public DataMemberAccessor(
        MemberInfo memberInfo,
        StorableAttribute autoStorableAttribute,
        object obj) {
      if (memberInfo.MemberType == MemberTypes.Field) {
        FieldInfo fieldInfo = (FieldInfo)memberInfo;
        this.getter = () => fieldInfo.GetValue(obj);
        this.setter = (value) => fieldInfo.SetValue(obj, value);
        this.type = fieldInfo.FieldType;
      } else if (memberInfo.MemberType == MemberTypes.Property) {
        PropertyInfo propertyInfo = (PropertyInfo)memberInfo;
        if (!propertyInfo.CanRead || !propertyInfo.CanWrite) {
          throw new NotSupportedException(
            "Storable properties must implement both a Get and a Set accessor. ");
        }
        this.getter = () => propertyInfo.GetValue(obj, null);
        this.setter = (value) => propertyInfo.SetValue(obj, value, null);
        this.type = propertyInfo.PropertyType;
      } else {
        throw new NotSupportedException(
                "The Storable attribute can only be applied to fields and properties.");
      }      
      this.name = autoStorableAttribute.Name != null ? autoStorableAttribute.Name : memberInfo.Name;      
      this.defaultValue = autoStorableAttribute.DefaultValue;                  
    }

    public DataMemberAccessor(
        string name, Type type, object defaultValue,
        Getter getter, Setter setter) {
      this.name = name;
      this.type = type;
      this.defaultValue = defaultValue;
      this.getter = getter;
      this.setter = setter;
    }

    public DataMemberAccessor(object o) {
      this.name = "";
      this.type = o.GetType();
      this.defaultValue = null;
      this.getter = () => o;
      this.setter = null;
    }

    public string Name {
      get { return this.name; }
    }

    public Type Type {
      get { return this.type; }
    }

    public object DefaultValue {
      get { return this.defaultValue; }
    }

    public object Get() {
      return this.getter();
    }

    public void Set(object value) {
      this.setter(value);
    }

    public override string ToString() {
      return String.Format("DataMember({0}, {1}, {2}, {3}, {4}",
        this.name,
        this.type == null ? "<null>" : this.type.FullName,
        this.defaultValue == null ? "<null>" : this.defaultValue,
        this.getter.Method,
        this.setter.Method);
    }
  }

}
